/* ********************************************************************
 * Copyright (c) 2019 Biological Modeling and Visualization Laboratory,
 * University of Calgary. All rights reserved.
 * ********************************************************************/

#include <cmath>
#include <algorithm>
#include <lpfgall.h>

#define DISPLAY_STEP 10

// constants
const double RadToDeg = 57.29577951;
const double TwoPi = 6.283185307;
const double PHI = 0.61803398875;

const double k1 = 0.02266; // lateral displacement parameter for young primordia

const double DAGE = 1.0; // age increment per time step

const double tNorm = 1600.0; // normalizing time factor in calls to func()

// minimum threshold distance for insertion
double thresholdDistance;
const double initialThresholdDistance = 3.5*M_PI;
#define USER_FUNC_TH_DIST(t) func(thDistNormalized,t/tNorm)

// active ring
const double DPHI = 0.0001; // angular step for scanning intervals, in radians
const double maxRingRadius = 12.4354;
#define USER_FUNC_AR_RADIUS(t) func(activeRingPos,t/tNorm) * maxRingRadius
#define ELLIPTIC_RING 1
#define ELLIPTIC_RING_FUNC(dr,ar,phi) 0.08*std::min(dr-ar,ar)*std::sin(2.*phi-50.0/RadToDeg)

// simulation time
const double dProgress = 0.1;   // time step increment per simulation step (dt)
const double EPS = 1e-5;
double Progress;               // progress of development (from 0 to 1)

const double TRANSITION_TIME_TO_34 = 660.;
const int NUM_BRACTS = 89; // number of bract primordia in this model


// declaration of a primordium 

struct primordium{
	double rho;	// primordium radius
	double phi;	// angular coordinate
	double r;	// radial cordinate
	V3d pos;	// Cartesian position
	double age;
	int neighbour_left;
	int neighbour_right;
	double inserted_dist;
};

primordium P[1000];		// array of primordia centers formed so far
primordium tp;	// trial/incipient primordium

int n;		// number of primordia formed so far;


// Find Cartesian coordinates, given radius and angular coordinate p.phi
void SyncCoordinates(primordium *p)
{
	p->pos.x = p->r*cos(p->phi); // convert to cartesian coordinates
	p->pos.z = p->r*sin(p->phi);
	p->pos.y = 0.0;
	
	return;
}

double PrimordiaDistance (const primordium& P1, const primordium& P2, double C_arclen, double C_radius)
{
    // compute arc length between primordia using the smallest angle between them
	double angle = std::fabs(P1.phi - P2.phi);
    angle = std::min(angle,TwoPi-angle);
    double arcLen = (P1.r+P2.r)*0.5 * angle;

	// measure distance as: sqrt( arclenth^2 + dRadius^2)
	double dR = P1.r - P2.r;
	double dist = std::sqrt(C_arclen*arcLen*arcLen + C_radius*dR*dR);

	return dist;
}

// Check if test primordium tp collides with previous primordia 

int HofmeisterTest(primordium *tp)
{
	extern int n;
	double min_dist1 = 1e11;
	double min_dist2 = 1e11+1.0;
	double min_dist3 = 1e11+2.0;
	int min_i1 = n+1;
	int min_i2 = n+2;
	int min_i3 = n+3;

	for (int i = 0; i<n; i++) {	// iterate over primordia
			double dist = PrimordiaDistance(*tp,P[i],1.,1.);
            if (dist < thresholdDistance)
				return 1;

			// check if this is the closest primordium
			if (dist < min_dist1) {
				min_dist3 = min_dist2;
				min_i3 = min_i2;
				min_dist2 = min_dist1;
				min_i2 = min_i1;
				min_dist1 = dist;
				min_i1 = i;
			} // else check if this is the second closest
			else if (dist < min_dist2) {
				min_dist3 = min_dist2;
				min_i3 = min_i2;
				min_dist2 = dist;
				min_i2 = i;
			}
			else if (dist < min_dist3) {
				min_dist3 = dist;
				min_i3 = i;
			}
	}

    // if there is one primordium or none, the min indicies should all be zero
    // this simplifies assignment of neighbours below
    if (n < 2) {
        min_i1 = min_i2 = min_i3 = 0;
    } 
    else {
        // if the three distances are almost the same,
        // the incipient primordium is tangential to three existing primordia
		if (std::fabs(min_dist2 - min_dist3) < 1e-5 &&
			std::fabs(min_dist1 - min_dist2) < 1e-5 && 
			std::fabs(min_dist1 - min_dist3) < 1e-5) {

            // pick the two youngest
			if (P[min_i1].age > P[min_i2].age && P[min_i1].age > P[min_i3].age) { // min_i1 is oldest
				min_i1 = min_i3;
			}
			else if (P[min_i2].age > P[min_i1].age && P[min_i2].age > P[min_i3].age) { // min_i2 is oldest
				min_i2 = min_i3;
			}
			//else { // min_i3 is oldest
			//}
		}
	}


    // save index of left and right neighbours (anti-clockwise and clockwise, respectively)
    // based on the angle between them
    double angle1 = P[min_i1].phi - tp->phi;
    angle1 -= TwoPi * std::floor(angle1/TwoPi); // wrap the angle from 0 to TwoPi

    // if the angle >= pi, min1 is CCW neighbour
    if (angle1 >= M_PI) {

        tp->neighbour_left = min_i1;
        P[min_i1].neighbour_right = n;

        tp->neighbour_right = min_i2;
        P[min_i2].neighbour_left = n;

    } else { // else min2 is CCW neighbour

        tp->neighbour_left = min_i2;
        P[min_i2].neighbour_right = n;

        tp->neighbour_right = min_i1;
        P[min_i1].neighbour_left = n;
    }

	return 0;	// if all primordia were far enough, collision has not occurred
}


module Draw();  // draws everything
module DrawReceptacle();
module Background();


Axiom: SB Background EB Draw() ;

// Initializations

int Phase = 1; 
double TransitionCount = 0;
const double TransitionTime = 200.0;
double minDist, dist;


Start:
{
	n = 0;			// there is no initial primordium
	Progress = 0.0;	

	thresholdDistance = initialThresholdDistance * USER_FUNC_TH_DIST(Progress);

	// trial primordium on active ring
    tp.r = USER_FUNC_AR_RADIUS(Progress);
	tp.rho = 0.;
	tp.age = 0.;
	tp.neighbour_left = -1; 
	tp.neighbour_right = -1; 
	tp.inserted_dist = thresholdDistance;

    // after rewind, temporarily reset receptacle radius
    P[0].r = tp.r;
    minDist = 0.;
    dist = 0.;

    Phase = 1;
    TransitionCount = 0;
}

StartEach: {

if (Phase == 1) {
	// Advance position of primordia
	for (int i=0; i<n; i++) {
		// advance primordia age
		P[i].age += DAGE * dProgress;

		// advance primordia centers
        P[i].r = P[i].r +
          tp.r * (std::pow(1.+PHI, 1.0/200.0) - 1.) * func(receptacleGrowth,Progress/tNorm) * dProgress;

        // for oldest primordium (defining radius of receptacle), compute distance to active ring
        if (i == 0) {
            dist = P[0].r - tp.r;
        }
        // if the distance is decreasing (from last saved distance), move all primordia to maintain the saved distance
        if (dist < minDist) {
            P[i].r = P[i].r + (minDist - dist);
        } else {
            minDist = dist; // update minimum distance
        }

		// recompute radius
		if (n < NUM_BRACTS) {
			P[i].inserted_dist = thresholdDistance;
        }
		P[i].rho = P[i].r * std::sin(P[i].inserted_dist*0.5/P[i].r); 
	}

	// Apply lateral displacement (right is clockwise, left is counter clockwise)
	for (int i = 0; i < n; i++) {

		int left_i = P[i].neighbour_left;
		int right_i = P[i].neighbour_right;

        // only incipient primordium can move laterally
        // identified by being younger than both of its neighbours
        if (n > 1 && P[i].r < 1.5*tp.r && P[i].age <= P[left_i].age && P[i].age <= P[right_i].age) {

            double dPhi = 0.;

            if (P[left_i].age >= P[right_i].age) { // CCW is older

                double arcLenL = PrimordiaDistance(P[i],P[left_i],1.,0.);

                double radiusL = 0.5*(P[i].r+P[left_i].r);

                double d = PrimordiaDistance(P[i],P[left_i],1.,1.);
                double a = arcLenL / d;

                // -k1*(-1) so lateral displacement is in CCW direction
                dPhi =  a * k1 * (arcLenL - a * thresholdDistance) / radiusL;

            } else { // CW is older

                double arcLenR = PrimordiaDistance(P[i],P[right_i],1.,0.);

                double radiusR = 0.5*(P[i].r+P[right_i].r);

                double d = PrimordiaDistance(P[i],P[right_i],1.,1.);
                double a = arcLenR / d;

                dPhi = - a * k1 * (arcLenR - a * thresholdDistance) / radiusR;
            }

            // error checking: the lateral displacement should not be too large
            if (std::abs(dPhi) > 0.1) {
                Printf ("Warning: change in phi is > 0.1. P[%d]: %g\n", i+1, std::abs(dPhi));
                PauseSimulation();
            }

            // update the angle
            double newPhi = P[i].phi - dPhi * dProgress;
            newPhi -= TwoPi * std::floor(newPhi/TwoPi); // wrap newPhi from 0 to TwoPi
  
            P[i].phi = newPhi;
        }

		SyncCoordinates(&P[i]);
	}

    double activeRingRadius = tp.r;
	double age = 1.; 
    if (n == 0) // in Model 1, the age of first primordium is 100
      age = 100.;

	// Insert new primordium if they do not collide with previous ones
	// Iterate over azimuths
	for (double phi = M_PI; phi < M_PI+TwoPi; phi += DPHI) {
		tp.phi = phi;
		if (tp.phi > TwoPi)
			tp.phi -= TwoPi;

#if ELLIPTIC_RING==1
        tp.r = activeRingRadius + ELLIPTIC_RING_FUNC(P[0].r,activeRingRadius,tp.phi);
#endif

		if (HofmeisterTest(&tp) == 0) {
			Printf("Producing primordium %d\n", n+1);

			tp.age = age;
			age += dProgress * DAGE;

			tp.inserted_dist = thresholdDistance;

		    tp.rho = tp.r * std::sin(thresholdDistance*0.5/tp.r); 

			P[n] = tp;
    		SyncCoordinates(&P[n]);
			++n;
		}
	}

	// Advance position of the active ring
    tp.r = USER_FUNC_AR_RADIUS(Progress);

	// stop simulation when the active ring has advanced to centre of receptacle
	if (tp.r < 0.05) { 
		Printf ("Simulation done.\n");
		Stop();
	}

	// Reduce the threshold distance for insertion of new primordia
	thresholdDistance = initialThresholdDistance * USER_FUNC_TH_DIST(Progress);

	// Advance time
	Progress += dProgress;

	if (StepNo() % DISPLAY_STEP*int(1./dProgress) == 0) {

		DisplayFrame();

        // transition at 34
        if (std::abs(Progress-TRANSITION_TIME_TO_34) < EPS) {
            Phase = 2;
            TransitionCount = 0;
        }
	}
} else if (Phase == 2) {

        TransitionCount += dProgress;

        if (TransitionCount >= TransitionTime)
            Phase = 1;

        if (StepNo() % DISPLAY_STEP*int(1./dProgress) == 0) {
            DisplayFrame();
        }

} else {
        Printf ("Unknown Phase\n");
}
}

derivation length: 1;
//derivation length: 5900; // 21P
//derivation length: 6600; // 34P
//derivation length: 690; // 55P

interpretation:
maximum depth: 2;

Background() :
{
    const double SF = 8.5;

	nproduce  MoveTo(0.,0.5,0.) SetCoordinateSystem(1.) Down(90) Left(90);
	nproduce SetColor(1);

    if (Phase == 2) {

        int color = std::min(143,117 + int( TransitionCount/TransitionTime * 26.0 ));
        produce SetColor(color) MoveTo(-0.25,0.,0.1) Left(-1) CurrentTexture(7) Surface(0,SF);

    } else {
        if (TransitionCount >= TransitionTime)
            produce CurrentTexture(-1);
    if (Progress + EPS >= 660)
        produce;
    if (Progress + EPS >= 612)
        produce MoveTo(-0.25,0.,0.1) Left(-1) CurrentTexture(7) Surface(0,SF);
    if (Progress + EPS >= 528.)
		produce MoveTo(-0.2,0.,0.2) Left(-1) CurrentTexture(6) Surface (0, SF);
    if (Progress + EPS >= 440.)
		produce MoveTo(-0.5,0.,-0.3) Left(0.5) CurrentTexture(5) Surface (0, SF);
    if (Progress + EPS >= 348.)
		produce MoveTo(0.05,0.,0.0) Left(4) CurrentTexture(4) Surface (0, SF);
    if (Progress + EPS >= 252.)
		produce MoveTo(0.,0.,-0.1) Left(-4) CurrentTexture(3) Surface (0, SF);
    if (Progress + EPS >= 180.)
		produce MoveTo(0.1,0.,-0.8) Left(5) CurrentTexture(2) Surface (0, SF);
    if (Progress + EPS >= 90.)
		produce MoveTo(0.,0.,-0.4) Left(-2) CurrentTexture(1) Surface (0, SF);
	produce MoveTo(-0.25,0.,-0.3) Right(0) CurrentTexture(0) Surface (0, SF);

    }
}

const double RING_WIDTH_SCALE = 5.*0.13; 

Draw():
{
	// Draw receptacle
	nproduce SetColor(15) SB MoveTo(0.,0.,0.) SetWidth(0.04)
             SetCoordinateSystem(1) ContourSides(64) DrawReceptacle EB;

	// Draw active ring
	if (tp.r >= 0.) {
		primordium ar = tp;

        // move to start position
    	ar.phi = M_PI;
#if ELLIPTIC_RING==1
        ar.r = tp.r + ELLIPTIC_RING_FUNC(P[0].r,tp.r,M_PI);
#endif
		SyncCoordinates(&ar);
		nproduce SB SetColor(254) SetWidth(RING_WIDTH_SCALE*0.5)
                    Up(90) Right(180) MoveTo3d(ar.pos+V3d(0.,0.0*RING_WIDTH_SCALE,0.));

        // draw many segements
        const double dphi = M_PI*0.01;
        //int i = 0;
		for (double phi = M_PI+dphi; phi < M_PI+TwoPi+dphi; phi += dphi) {
			ar.phi = phi;
#if ELLIPTIC_RING==1
            ar.r = tp.r + ELLIPTIC_RING_FUNC(P[0].r,tp.r,ar.phi);
#endif
			SyncCoordinates(&ar);
			nproduce LineTo3d(ar.pos+V3d(0.,0.0*RING_WIDTH_SCALE,0.));
		}
		nproduce EB;
  	}

	// Draw primordia
	int color = 0;
    nproduce SB Up(90) SetWidth(0.08);
	for (int i=0; i<n; i++) {
        // Phase == 2 : transistions between visualization as circles and spheres
        if (Phase == 2) {
             nproduce MoveTo3d(P[i].pos+V3d(0.,-0.1,0.));
             color = int( TransitionCount/TransitionTime * 64.0 ) ;
             nproduce SetColor((64-color)+144) SB Down(90) Sphere(P[i].rho) EB;
             nproduce SetColor(color+144) SB CircleB(P[i].rho) EB;

        } else { 
            // Phase == 1: draw primordia as circles, and then later as spheres
            if (TransitionCount < TransitionTime) {
                double a = P[i].age;
                if (a>250)
                    a = 250;
                nproduce SB SetWidth(0.085*1.5) SetColor(7) LineTo3d(P[i].pos+V3d(0.,-0.1,0.)) // draw "spoke"
                        SetColor(16+ int(a/2.5)) SB Up(90) f(1.0) Down(90)  Circle(0.6-0.085*0.5)
                                                    Up(90) f(1.0) Down(90) SetColor(2) Circle(0.6-0.085*0.5)
                                                    SetColor(9)  CircleB(0.6) EB EB;	// draw primordium
            } else {
                color = 8;
                if (i >= NUM_BRACTS) {
                    color = 9;
                }
                nproduce MoveTo3d(P[i].pos+V3d(0.,-0.1,0.)) SetColor(color) Sphere(P[i].rho);
            }
        }
    }
    nproduce EB;


	produce;
}

DrawReceptacle: 
{
    if (Phase == 2) {
        int colour = std::min(143,117 + int( TransitionCount/TransitionTime * 26.0 )) ;
    	produce SB Up(90) SetWidth(RING_WIDTH_SCALE*0.25) SetColor(colour) CircleB(P[0].r) EB;
    } else {
        if (TransitionCount < TransitionTime) {
	        produce SB Up(90) SetWidth(RING_WIDTH_SCALE*0.25) SetColor(3) CircleB(P[0].r) EB;
        }
    }
}


